import lxml.etree
import os
import shutil
import sys
import time
import tempfile
import paramiko
import io
import socket


class OscapWrapper(object):
    NAMESPACES = {
    'common': 'http://oval.mitre.org/XMLSchema/oval-common-5',
    'definitions': 'http://oval.mitre.org/XMLSchema/oval-definitions-5',
    'results': 'http://oval.mitre.org/XMLSchema/oval-results-5',
    'linux': 'http://oval.mitre.org/XMLSchema/oval-definitions-5#linux',
    'unix': 'http://oval.mitre.org/XMLSchema/oval-definitions-5#unix',
    'ind': 'http://oval.mitre.org/XMLSchema/oval-definitions-5#independent',
    'info': 'http://oval.mitre.org/XMLSchema/oval-system-characteristics-5',
    }

    def __init__(self, target, localoval, vendoroval=None,
                 username='root', key=None, keypass=None,
                 port=22, tempdir=None):
        self.target = target
        self.localoval = localoval
        self.vendoroval = vendoroval
        self.username = username
        self.port = port
        self.key = key
        self.keypass = keypass
        self.tempdir = tempdir
        if isinstance(self.tempdir, type(None)):
            self.tempdir = tempfile.mkdtemp(prefix='oscapwrap')
        self.merge = False if isinstance(vendoroval, type(None)) else True
        self.client = None
        if not self._check_params():
            raise Exception('Invalid params')
        self._prepare_ssh()

    def _check_params(self):
        if isinstance(self.username, type(None)):
            return False
        if not isinstance(self.vendoroval, type(None)):
            if not os.path.exists(self.vendoroval):
                return False
        if not os.path.exists(self.localoval):
            return False
        return True

    def _prepare_ssh(self):
        self.client = paramiko.SSHClient()
        self.client.load_system_host_keys()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if isinstance(self.key, str):
            self.key = io.StringIO(self.key)
        if not isinstance(self.key, type(None)):
            self.key = paramiko.RSAKey.from_private_key(
                        self.key, password=self.keypass)
        self.client.connect(self.target, port=self.port,
                            username=self.username,
                            pkey=self.key, timeout=10.0,
                            banner_timeout=10.0)
        self.sftp = self.client.open_sftp()

    def _RunCommand(self, cmd):
        session = self.client.get_transport().open_session()
        session.exec_command(cmd)
        res = session.recv_exit_status()
        session.close()
        return res

    def _RunOscap(self, inpfile, resfile):
        cmd = 'oscap oval eval --results {0} {1}'.format(resfile, inpfile)
        return self._RunCommand(cmd)

    def _CopyOvalDefs(self, loval, rtmpdir):
        roval = '/'.join([rtmpdir, os.path.basename(loval)])
        self.sftp.put(loval, roval)

    def _CopyScanResults(self, rres, keep=True):
        if not os.path.exists(self.tempdir):
            os.mkdir(self.tempdir)
        lres = '/'.join([self.tempdir, os.path.basename(rres)])
        self.sftp.get(rres, lres)
        if not keep:
            self._RunCommand('rm -rf {0}'.format(rtmpdir))

    def _MergeResults(self, out):
        vendorxml = '/'.join([self.tempdir, 'vendor-results.xml'])
        localxml = '/'.join([self.tempdir, 'local-results.xml'])
        fids = dict()
        tids = dict()
        toreplace = list()
        for definition in vendorxml.xpath('.//*[@class="vulnerability"]'):
            did = definition.attrib['id']
            cve = [x.attrib['ref_id'] for x in
                    definition.xpath('.//*[@source="CVE"]')][0]
            fids[cve] = did
        for definition in localxml.xpath('.//*[@class="vulnerability"]'):
            did = definition.attrib['id']
            cve = [x.attrib['ref_id'] for x in
                    definition.xpath('.//*[@source="CVE"]')][0]
            tids[cve] = did

        toreplace = [fids[x] for x in fids.keys() if x in tids]

        for res in vendorxml.xpath('.//results:definition', namespaces=NAMESPACES):
            if res.attrib['definition_id'] in toreplace:
                for trueres in res.xpath('.//*[@result="true"]'):
                    trueres.set('result', 'false')

        ftests = vendorxml.xpath('.//definitions:tests',
                                 namespaces=NAMESPACES)
        fobjects = vendorxml.xpath('.//definitions:objects',
                                 namespaces=NAMESPACES)
        fdefs = vendorxml.xpath('.//definitions:definitions',
                                 namespaces=NAMESPACES)
        fstates = vendorxml.xpath('.//definitions:states',
                                 namespaces=NAMESPACES)

        fresdefs = vendorxml.xpath('.//results:definitions',
                                 namespaces=NAMESPACES)

        for tests in localxml.xpath('.//definitions:tests',
                                    namespaces=NAMESPACES):
            for test in tests: ftests[0].append(test)

        for objs in localxml.xpath('.//definitions:objects',
                                   namespaces=NAMESPACES):
            for obj in objs: fobjects[0].append(obj)

        for stes in localxml.xpath('.//definitions:states',
                                   namespaces=NAMESPACES):
            for ste in stes: fstates[0].append(ste)

        xids = list()
        for xdef in localxml.xpath('.//*[@class="vulnerability"]'):
            cve = xdef.xpath('.//*[@source="CVE"]')[0].attrib['ref_id']
            if not cve in fids:
                fdefs[0].append(xdef)
                xids.append(xdef.attrib['id'])

        for xdef in localxml.xpath('.//results:definitions',
                                   namespaces=NAMESPACES)[0]:
            if xdef.attrib['definition_id'] in xids:
                fresdefs[0].append(xdef)

        ovaldef = vendorxml.xpath('.//definitions:oval_definitions',
                                  namespaces=NAMESPACES)[0]
        name = ovaldef.xpath('.//common:product_name', namespaces=NAMESPACES)[0]
        ts = ovaldef.xpath('.//common:timestamp', namespaces=NAMESPACES)[0]
        name.text = 'Mirantis merged OVAL Generator'
        ts.text = time.strftime('%Y-%m-%dT%H:%M:%S')
        with open(out, 'w') as out:
            out.write(lxml.etree.tostring(vendorxml))

    def scan(self):
        if self.merge:
            out = '/'.join([self.tempdir, 'result.xml'])
        else:
            out = '/'.join([self.tempdir, 'local-results.xml'])
        rtmpdir = '/tmp/oscap-{0}'.format(str(int(time.time())))
        self._RunCommand('mkdir -p {0}'.format(rtmpdir))
        self._CopyOvalDefs(self.localoval, rtmpdir)
        if self.merge:
            self._CopyOvalDefs(self.vendoroval, rtmpdir)
            rvendoroval = '/'.join([rtmpdir,
                                    os.path.basename(self.vendoroval)])
            rvendorres = '/'.join([rtmpdir, 'vendor-results.xml'])
        rlocaloval = '/'.join([rtmpdir, os.path.basename(self.localoval)])
        rlocalres = '/'.join([rtmpdir, 'local-results.xml'])
        if self.merge:
            if not self._RunOscap(rvendoroval, rvendorres) == 0:
                raise Exception('Vendor oval scan failed')
        if not self._RunOscap(rlocaloval, rlocalres) == 0:
            raise Exception('Local oval scan failed')
        self._CopyScanResults(rlocalres)
        if self.merge:
            self._CopyScanResults(rvendorres, keep=False)
            self._MergeResults(out)
        else:
            self._RunCommand('rm -rf {0}'.format(rtmpdir))
        return out
        #shutil.rmtree(self.tempdir)

