import json
import time
import socket
import ssl
import lxml.etree
import io
import datetime
import base64


class OpenVas(object):
    DEFAULT_PROFILE = 'Full and very deep ultimate'
    DEFAULT_REPORT_FORMAT = 'CVE Results'
    AVAILABLE_GET_COMMANDS = ('alerts', 'configs', 'credentials', 'feeds',
                              'port_lists', 'reports', 'report_formats',
                              'results', 'scanners', 'schedules', 'settings',
                              'tags', 'targets', 'tasks')
    AVAILABLE_CREATE_COMMANDS = ('alert', 'credential', 'filter', 'group',
                                 'note', 'port_list', 'port_range', 'report',
                                 'report_format', 'tag', 'target', 'task')

    def __init__(self, username='admin', passwd='admin',
                 host='127.0.0.1', port=9390,
                 sshuser='root', sshkey=None, sshkeypass=None):
        self.username = username
        self.passwd = passwd
        self.host = host
        self.port = port
        self.socket = None
        self.sshuser = sshuser
        self.sshkey = sshkey
        self.sshkeypass = sshkeypass
        self.cbhost = '127.0.0.1'
        self.cbport = 54321

    def open(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket = sock = ssl.wrap_socket(sock)
        sock.connect((self.host, self.port))
        self._authenticate()

    def close(self):
        self.socket.close()
        self.socket = None

    def reconnect(self):
        self.close()
        self.open()

    def _authenticate(self):
        data = self.Dict2XML('authenticate', [
                             {'credentials': [
                                 {'username': self.username},
                                 {'password': self.passwd}
                             ]}])
        out = lxml.etree.fromstring(self._RunXmlCommand(
                                    lxml.etree.tostring(data)))
        assert int(out.attrib['status']) == 200
        return True

    def Dict2XML(self, tag, data):
        elem = lxml.etree.Element(tag)
        if isinstance(data, str):
            elem.text = data
            return elem
        elif isinstance(data, int):
            elem.text = str(data)
            return elem
        elif isinstance(data, type(None)):
            return elem
        for item in data:
            if isinstance(item, dict):
                for key, value in item.items():
                    if key == '@':
                        if isinstance(value, dict):
                            for attr_name, attr_value in value.items():
                                elem.attrib[attr_name] = attr_value
                        elif isinstance(value, list):
                            for attr_item in value:
                                for attr_name, attr_value in attr_item.items():
                                    elem.attrib[attr_name] = attr_value
                    elif key.startswith('@'):
                        elem.attrib[key[1:]] = value
                    else:
                        elem.append(self.Dict2XML(key, value))
            elif isinstance(item, list):
                for x in item:
                    if isinstance(x, dict):
                        for k, v in x.iteritems():
                            elem.append(self.Dict2XML(k, v))
            elif isinstance(item, str):
                elem.text = item
            else:
                elem.append(lxml.etree.Element(item))
        return elem

    def _RunXmlCommand(self, xmlstring):
        BLOCK_SIZE = 1024
        response = str()
        if isinstance(xmlstring, str):
            xmlstring = xmlstring.encode('utf-8')
        self.socket.send(xmlstring)
        while 1:
            res = self.socket.recv(BLOCK_SIZE)
            response += res.decode('utf-8')
            if len(res) < BLOCK_SIZE:
                break
        return response

    def _GetCmd(self, key):
        assert len(key.split()) == 1, 'You can use only one command'
        assert key in self.AVAILABLE_GET_COMMANDS,\
            "Unknown or disabled command 'get_{0}'".format(key)
        out = self._RunXmlCommand("<get_{0}/>".format(key))
        return lxml.etree.fromstring(out)

    def _CreateCmd(self, key, params):
        key = key.replace('create_', '') if key.startswith('create_') else key
        assert key in self.AVAILABLE_CREATE_COMMANDS,\
            "Unknown or disabled command 'create_{}'".format(key)
        xml = self.Dict2XML("create_{0}".format(key), params)
        out = self._RunXmlCommand(lxml.etree.tostring(xml))
        return lxml.etree.fromstring(out)

    def _RawXmlCommand(self, key, params):
        xml = self.Dict2XML(key, params)
        return self._RunXmlCommand(lxml.etree.tostring(xml))

    def _CreateTarget(self, address, cred, ports, label=None):
        label = label if label else address
        conf = [
            {'name': label},
            {'hosts': address},
            {'ssh_lsc_credential': [
                {'port': 22},
                {'@id': cred},
            ]}
        ]
        if not (ports.startswith('U:') or ports.startswith('T:')):
            conf.append({'port_list': [{'@id': ports}]})
        else:
            conf.append({'port_range': ports})
        target = self._CreateCmd('target', conf)
        assert int(target.attrib['status']) == 201
        return target.attrib['id']

    def _GetOrCreateCredentialId(self, name):
        creds = self._GetCmd('credentials')
        assert int(creds.attrib['status']) == 200
        creds = creds.xpath('./credential')
        for cred in creds:
            xname = cred.find('./name').text
            if not isinstance(name, type(None)) and name == xname:
                return cred.attrib['id']
        key = [{'private': self.sshkey}]
        if not isinstance(self.sshkeypass, type(None)):
            key.append({'phrase': self.sshkeypass})
        conf = [
            {'name': name},
            {'comment': 'Mirantis default credentials'},
            {'key': key},
            {'login': self.sshuser},
            {'type': 'usk'} # Usename Secret Key
            ]
        res = self._CreateCmd('credential', conf)
        assert int(res.attrib['status']) == 201, "Credential create failed"
        return res.attrib['id']

    def _GetTargets(self):
        tgts = dict()
        targets = self._GetCmd('targets')
        assert int(targets.attrib['status']) == 200
        for tgt in targets.xpath('./target'):
            xname = tgt.find('./name').text
            tgts[xname] = tgt.attrib['id']
        return tgts

    def _GetTargetId(self, target):
        targets = self._GetCmd('targets')
        assert int(targets.attrib['status']) == 200
        for tgt in targets.xpath('./target'):
            if tgt.find('./name').text == target:
                return tgt.attrib['id']
        return None

    def _CreateTask(self, targetid, configid, name, alertid):
        conf = [
            {'name': name},
            {'target': [{'@id': targetid}]},
            {'config': [{'@id': configid}]},
            {'alert': [{'@id': alertid}]},
        ]
        out = self._CreateCmd('task', conf)
        assert int(out.attrib['status']) == 201
        return out.attrib['id']

    def _CreateContainer(self, name):
        conf = [
            {'target': [{'@id': '0'}]},
            {'name': name},
            {'comment': 'Merged reports container'},
        ]
        out = self._CreateCmd('task', conf)
        assert int(out.attrib['status']) == 201
        return out.attrib['id']

    def _StartTask(self, taskid):
        xml = '<start_task task_id="{}"/>'.format(taskid)
        report = lxml.etree.fromstring(self._RunXmlCommand(xml))
        assert int(report.attrib['status']) == 202
        return report.attrib['status_text']

    def GetTaskStatus(self, taskid):
        res = dict({'status': None, 'progress': 100})
        tasks = self._GetCmd('tasks')
        assert int(tasks.attrib['status']) == 200
        for task in tasks.xpath('./task'):
            if task.attrib['id'] == taskid:
                if task.find('./status').text == 'Running':
                    res['progress'] = task.find('./progress').text
                res['status'] = task.find('./status').text.lower()
        return res

    def GetTaskList(self, active=False):
        tasks = self._GetCmd('tasks')
        assert int(tasks.attrib['status']) == 200
        res = dict()
        for task in tasks.xpath('./task'):
            if active:
                if task.find('./status').text == 'Running':
                    if not task.find('./target').attrib['id'] == str():
                        res[task.attrib['id']] = task.find('./name').text
            else:
                if not task.find('./target').attrib['id'] == str():
                    res[task.attrib['id']] = task.find('./name').text
        return res

    def _GetPorts(self):
        _list = dict()
        ports = self._GetCmd('port_lists')
        assert int(ports.attrib['status']) == 200
        for port in ports.xpath('./port_list'):
            xname = port.find('./name').text
            _list[xname] = port.attrib['id']
        return _list

    def _GetAllTasks(self):
        taskids = list()
        tasks = self._GetCmd('tasks')
        assert int(tasks.attrib['status']) == 200
        for task in tasks.xpath('./task'):
            taskids.append(task.attrib['id'])
        return taskids

    def _GetFilteredTasks(self, count=10, status='Done'):
        taskdict = dict()
        res = list()
        tasks = lxml.etree.fromstring(
                self._RawXmlCommand('get_tasks', [
                {'@filter': 'status={0}'.format(status)}]))
        for task in tasks.xpath('./task'):
            merged = True
            info = list()
            if task.find('./target').attrib['id'] == str():
                continue # Skip containers
            if status == 'Done':
                ts = task.find('./last_report/report/scan_end').text
                rid = task.find('./comment').text
                if isinstance(rid, type(None)):
                    rid = task.find('./last_report/report').attrib['id']
                    merged = False
                _status = rid
            else:
                ts = task.find('./modification_time').text
                _status = task.find('progress').text
            ts = ts.replace('Z', '+0000')
            ts = datetime.datetime.strptime(ts, '%Y-%m-%dT%H:%M:%S%z')
            ts = str(int(ts.timestamp()))
            name = task.find('name').text
            tid = task.attrib['id']
            if ts in taskdict:
                ts = ts+'_'
            taskdict[ts] = [name, tid, _status, merged]
        for x,y in sorted(taskdict.items())[:count]:
            res.append(y)
        return res

    def _GetReportId(self, taskid):
        reports = self._GetCmd('reports')
        assert int(reports.attrib['status']) == 200
        for report in reports.xpath('./report'):
            if report.find('./task').attrib['id'] == taskid:
                return report.attrib['id']
        return None

    def _GetTaskId(self, targetid):
        tasks = self._GetCmd('tasks')
        assert int(tasks.attrib['status']) == 200
        for task in tasks.xpath('./task'):
            xname = task.find('./name').text
            xstatus = task.find('./status').text
            if xname == targetid and status == 'Running':
                return task.attrib['id']
        return None

    def _GetTidByName(self, taskname):
        tasks = self._GetCmd('tasks')
        assert int(tasks.attrib['status']) == 200
        for task in tasks.xpath('./task'):
            xname = task.find('./name').text
            if xname == taskname:
                return task.attrib['id']
        return None

    def _GetProfiles(self):
        res = dict()
        configs = self._GetCmd('configs')
        assert int(configs.attrib['status']) == 200
        for config in configs.xpath('./config'):
            xname = config.find('./name').text
            res[xname] = config.attrib['id']
        return res

    def RunScan(self, ip, profile='Full and very deep ultimate',
                ports='All TCP and Nmap 5.51 top 1000 UDP',
                credentials='root-ssh'):

        AVAILABLE_PROFILES = self._GetProfiles()
        AVAILABLE_PORTS = self._GetPorts()
        AVAILABLE_TARGETS = self._GetTargets()

        assert profile in AVAILABLE_PROFILES, "Unknown scan profile"
        profileid = AVAILABLE_PROFILES[profile]

        targetid = self._GetTargetId(ip)
        tasklabel = '{0} ({1}) - {2}'.format(
                    time.strftime('%Y-%m-%d %H:%M:%S'),
                    time.tzname[0],
                    ip)

        if ports in AVAILABLE_PORTS:
            ports = AVAILABLE_PORTS[ports]

        if isinstance(targetid, type(None)):
            credentialsid = self._GetOrCreateCredentialId(name=credentials)
            targetid = self._CreateTarget(ip, credentialsid, ports, label=ip)

        taskid = self._GetTaskId(ip)

        cbhost = '127.0.0.1' if self.cbhost == '0.0.0.0' else self.cbhost
        alertaddr = ':'.join([cbhost, str(self.cbport)])
        alerturi = 'http://{0}/report/merge/$n'.format(alertaddr)
        alertname = 'API Callback to {0}'.format(alertaddr)

        alertid = self._GetOrCreateAlert([
                        {'name': alertname},
                        {'condition': 'Always'},
                        {'event': ['Task run status changed',
                            {'data': [
                                {'name': 'status'},
                                'Done',
                            ]},
                        ]},
                        {'method': ['HTTP Get',
                            {'data': [
                                {'name': 'URL'},
                                alerturi,
                            ]}
                        ]}
                        ])

        if isinstance(taskid, type(None)):
            taskid = self._CreateTask(targetid, profileid, tasklabel, alertid)

        self._StartTask(taskid)
        time.sleep(3)
        status = self.GetTaskStatus(taskid)
        return dict({'id': taskid, 'status': status})

    def _GetReportFormats(self):
        res = dict()
        rfmts = self._GetCmd('report_formats')
        assert int(rfmts.attrib['status']) == 200
        for rfmt in rfmts.xpath('./report_format'):
            xname = rfmt.find('./name').text
            res[xname] = rfmt.attrib['id']
        return res

    def _GetReport(self, reportid, rformat='XML'):
        if rformat.lower() == 'csv':
            rformat = 'CSV Results'
        elif rformat.lower() == 'latex':
            rformat = 'LaTeX'

        AVAILABLE_REPORT_FORMATS = self._GetReportFormats()

        assert rformat in AVAILABLE_REPORT_FORMATS, "Unknown report format"
        formatid = AVAILABLE_REPORT_FORMATS[rformat]
        assert not isinstance(reportid, type(None)), "Report wasn't found"
        data = self._RawXmlCommand('get_reports', [{'@format_id': formatid},
                                             {'@report_id': reportid},
                                             {'@min_qod': '10'}])
        if rformat == 'XML':
            return data
        out = lxml.etree.fromstring(data)
        assert int(out.attrib['status']) == 200
        report = out.find('./report').find('./report_format').tail
        return base64.b64decode(report)

    def GetReport(self, taskid, rformat='XML'):
        reportid = self._GetReportId(taskid)
        return self._GetReport(reportid, rformat=rformat)

    def _UploadReport(self, xmlreport, container='Mirantis merged reports'):
        containerid = None
        tasks = self._GetCmd('tasks')
        for task in tasks.xpath('./task'):
            if task.find('./name').text == container:
                containerid = task.attrib['id']
                break
        if isinstance(containerid, type(None)):
            containerid = self._CreateContainer(container)
        body = '<create_report>{1}<task id="{0}"/></create_report>'.format(
                                                     containerid, xmlreport)
        resp = lxml.etree.fromstring(self._RunXmlCommand(body))
        assert int(resp.attrib['status']) == 201
        return resp.attrib['id']

    def _GetOrCreateAlert(self, params):
        for i in params:
            if i.get('name', False):
                name = i['name']
                break
        alerts = self._GetCmd('alerts')
        for alert in alerts.xpath('./alert'):
            if alert.find('./name').text == name:
                return alert.attrib['id']
        out = self._CreateCmd('alert', params)
        assert int(out.attrib['status']) == 201
        return out.attrib['id']

    def UpdateReport(self, taskid, closedcves):
        report = self.GetReport(taskid, rformat='XML')
        xml = lxml.etree.parse(io.StringIO(report))
        root = xml.getroot()
        count = 0
        for result in root.xpath('.//report/report/results/*'):
            parent = result.getparent()
            cveselem = result.xpath('./nvt/cve')[0]
            cves = cveselem.text.split(',')
            x = list(set(closedcves) & set(cves))
            if cves[0] == 'NOCVE':
                continue
            cves = list(set(cves) ^ set(x))
            if len(cves) == 0:
                count +=1
                parent.remove(result)
            elif len(cves) > 0:
                cveselem.text = ','.join(cves)
                count +=1
        report = root.find('./report')
        taskid = report.find('./task').attrib['id']
        nrepid = self._UploadReport(lxml.etree.tostring(report))
        return self._UpdateTaskComment(taskid, nrepid)

    def _UpdateTaskComment(self, taskid, text):
        resp = self._RawXmlCommand('modify_task', [
                                   {'@task_id': taskid},
                                   {'comment': text},
                                  ])
        resp = lxml.etree.fromstring(resp)
        assert int(resp.attrib['status']) == 200, 'Fail to update task comment'
        return True

    def GetCvesFromMergedReport(self, taskid):
        tasks = self._GetCmd('tasks')
        task = tasks.find('.//*[@id="{0}"]'.format(taskid))
        repid = task.find('./comment').text
        result = dict({'open': list(), 'closed': list()})
        if repid == '':
            return json.dumps(result)
        rep = lxml.etree.fromstring(self._GetReport(repid))
        report = rep.find('./report')
        for threat in report.xpath('.//result/threat'):
            if not threat.text == 'Log':
                for cve in threat.getparent().xpath('.//nvt/cve'):
                    result['open'].append(cve.text)
        for value in report.xpath('.//host/detail/value'):
            if value.text.startswith('CVE-'):
                result['closed'] += [n.strip() for n in value.text.split(',')]
        return json.dumps(result)
