import io
import os
import lxml.etree
import shutil
import urllib.request
from mirascan.openvas import OpenVas
from mirascan.openscap import OscapWrapper

class Scanner(OpenVas):

    def __init__(self, username='admin', passwd='admin',
                 host='127.0.0.1', port=9390, ovalpath='/root/',
                 savepath='static/reports', key=None):
        OpenVas.__init__(self, username=username, passwd=passwd,
                         host=host, port=port)
        self.passwd = passwd
        self.ovalpath = ovalpath
        self.savepath = savepath
        self.open()

    def __str__(self):
        return '{0} on {1}:{2} with pass {3}'.format(
                self.username,
                self.host,
                self.port,
                self.passwd)

    def _ScanOscap(self, target, localoval, vendoroval=None,
                   username=None, key=None, keypass=None):
        oscap = OscapWrapper(target, localoval, vendoroval=vendoroval,
                             username=username, key=key)
        return oscap.scan()

    def Scan(self, target):
        return self.RunScan(target)

    def UpdateResults(self, taskname, localoval, username=None, key=None):
        tid = self._GetTidByName(taskname)
        if not self.GetTaskStatus(tid)['status'] == 'done':
            return False
        ovasreport = self.GetReport(tid, rformat='XML')
        report = lxml.etree.fromstring(ovasreport)
        target = report.find('./report/report/results/result/host').text
        if isinstance(target, type(None)):
            return False
        reppath = self._ScanOscap(target,
                                  localoval,
                                  username=username,
                                  key=key)
        sreport = lxml.etree.parse(reppath)
        ids = dict()
        closedcves = list()
        opencves = list()
        for xdef in sreport.xpath('.//*[@class="vulnerability"]'):
            cveid = xdef.xpath('.//*[@source="CVE"]')[0].attrib['ref_id']
            for xres in sreport.xpath('.//*[@definition_id="{0}"]'.format(
                                                xdef.attrib['id'])):
                if xres.attrib['result'] == 'true':
                    opencves.append(cveid)
                else:
                    closedcves.append(cveid)
        del(sreport)
        self.UpdateReport(tid, closedcves)
        out = '/'.join([os.path.dirname(__file__), self.savepath])
        with open('/'.join([out, '{0}.xml'.format(tid)]), 'w') as res:
            res.write(self._MergeVasResults(open(reppath).read().encode(
                                                'utf-8'), ovasreport))
        shutil.rmtree(os.path.dirname(reppath))
        return True

    def __UpdateContent(self, contenttype):
        out = lxml.etree.fromstring(
                self._RunXmlCommand("<sync_{0}/>".format(contenttype)))
        return int(out.attrib['status']) == 202

    def UpdateData(self, scap=False, feed=False, cert=False,
                   oval=False, ovalurl=None, ovalpath=None):
        if scap:
            self.__UpdateContent('scap')
        if feed:
            self.__UpdateContent('feed')
        if cert:
            self.__UpdateContent('cert')
        if oval:
            if not ovalurl or not ovalpath:
                return False
            with open(ovalpath, 'w') as o:
                req = urllib.request.urlopen(ovalurl)
                if not req.code // 100 == 2:
                    return False
                o.write(req.read().decode('utf-8'))
        return True

    def _MergeVasResults(self, scapxml, vasxml):
        openscapdata = lxml.etree.fromstring(scapxml)
        openvasdata = lxml.etree.fromstring(vasxml)
        opencves = list()
        closedcves = list()
        idtocvemap = dict()
        for cve in openvasdata.xpath('./report/results/result/nvt/cve/*'):
            opencves.append(cve.text)
        opencves = list(set(opencves))
        for i in openvasdata.xpath('./report/host/detail'):
            if i.find('./name').text == 'Closed CVE':
                closedcves += [x.strip() for x in
                               i.find('./value').text.split(',')]
        closedcves = list(set(closedcves))
        for ovaldef in openscapdata.xpath('.//*[@class="vulnerability"]'):
            cveid = ovaldef.find('.//*[@source="CVE"]').get('ref_id')
            idtocvemap[ovaldef.get('id')] = cveid
        for xdef in openscapdata.find('.//results:definitions',
                                      namespaces=OscapWrapper.NAMESPACES):
            defid = xdef.get('definition_id')
            result = xdef.get('result')
            if defid not in idtocvemap:
                continue
            cveid = idtocvemap[defid]
            xdef.attrib['mirantis_result'] = result
            if result == 'true':
                if cveid in closedcves:
                    xdef.attrib['mirantis_result'] = 'missed'
            elif result == 'false':
                if cveid in opencves:
                    xdef.attrib['mirantis_result'] = 'fixed'
                    opencves.remove(cveid)
            elif result == 'unknown':
                if cveid in closedcves:
                    xdef.attrib['mirantis_result'] = 'false'
                elif cveid in opencves:
                    xdef.attrib['mirantis_result'] = 'true'
                    opencves.remove(cveid)
        for cveid in opencves:
            cid = ''.join(cveid.split('-')[1:])
            dtsotmpl = '{:0<%d}' % int(len(cid) + 4)
            dtsonum = dtsotmpl.format(cid)
            dist = 'com.mirantis'
            # Make a definition in oval definitions
            _def = lxml.etree.Element('definition')
            _def.attrib['id'] = 'oval:{0}:def:{1}'.format(dist, dtsonum)
            _def.attrib['class'] = 'vulnerability'
            _def.attrib['version'] = '1'
            meta = lxml.etree.SubElement(_def, 'metadata')
            title = lxml.etree.SubElement(meta, 'title')
            title.text = '{0} on Ubuntu 16.04'.format(cveid)
            reference = lxml.etree.SubElement(meta, 'reference')
            reference.attrib['source'] = 'CVE'
            reference.attrib['ref_id'] = '{0}'.format(cveid)
            reference.attrib['ref_url'] = \
                'https://nvd.nist.gov/vuln/detail/{0}'.format(cveid)
            criteria = lxml.etree.SubElement(_def, 'criteria')
            criterion = lxml.etree.SubElement(criteria, 'criterion')
            criterion.attrib['test_ref'] = \
                'oval:{0}:tst:{1}'.format(dist, dtsonum)
            openscapdata.find('.//definitions:definitions',
                              namespaces=OscapWrapper.NAMESPACES).append(_def)
            # Make a definition in oval results
            _def = lxml.etree.Element('definition')
            _def.attrib['definition_id'] = \
                'oval:{0}:def:{1}'.format(dist, dtsonum)
            _def.attrib['result'] = 'true'
            _def.attrib['version'] = '1'
            _def.attrib['mirantis_result'] = 'true'
            openscapdata.find('.//results:definitions',
                              namespaces=OscapWrapper.NAMESPACES).append(_def)
        return lxml.etree.tostring(openscapdata,
                                   pretty_print=True).decode('utf-8')

